@@ -1,3 +1,4 @@
+import functools
 import uuid
 from copy import deepcopy
 from datetime import datetime, timedelta
@@ -6,12 +7,13 @@ from typing import Union, Optional, Dict, Literal, List
 import aiofiles
 
 from glQiwiApi.abstracts import AbstractPaymentWrapper
-from glQiwiApi.api import HttpXParser
-from glQiwiApi.data import Response, Transaction, Identification, Limit, \
-    Bill, Commission
+from glQiwiApi.aiohttp_custom_api import CustomParser
+from glQiwiApi.data import Response, Identification
 from glQiwiApi.exceptions import InvalidData, InvalidCardNumber
+from glQiwiApi.data_models import QiwiAccountInfo, Transaction, Bill, BillError, Statistic, Limit, Account
+from glQiwiApi.data_models.basics import Sum, Commission
 from glQiwiApi.qiwi.basic_qiwi_config import *
-from glQiwiApi.utils import datetime_to_str_in_iso, DataFormatter
+from glQiwiApi.utils import datetime_to_str_in_iso, DataFormatter, multiply_objects_parse, simple_multiply_parse
 
 
 class QiwiWrapper(AbstractPaymentWrapper):
@@ -23,23 +25,35 @@ class QiwiWrapper(AbstractPaymentWrapper):
     """
 
     def __init__(self,
-                 api_access_token: Optional[str],
-                 phone_number: Optional[str],
+                 api_access_token: Optional[str] = None,
+                 phone_number: Optional[str] = None,
                  secret_p2p: Optional[str] = None,
-                 public_p2p: Optional[str] = None) -> None:
+                 public_p2p: Optional[str] = None,
+                 without_context: bool = False) -> None:
         """
         :param api_access_token: токен, полученный с https://qiwi.com/api
         :param phone_number: номер вашего телефона с +
         :param secret_p2p: секретный ключ, полученный с https://p2p.qiwi.com/
         :param public_p2p: публичный ключ, полученный с https://p2p.qiwi.com/
+        :param without_context: булевая переменная, указывающая будет ли объект класса "глобальной" переменной
+         или будет использована в async with контексте
         """
-        self._parser = HttpXParser()
+        self._parser = CustomParser(without_context=without_context, messages=ERROR_CODE_NUMBERS)
         self.api_access_token = api_access_token
         self.phone_number = phone_number
         self._formatter = DataFormatter()
         self.public_p2p = public_p2p
         self.secret_p2p = secret_p2p
 
+    async def __aenter__(self) -> 'QiwiWrapper':
+        """Создаем сессию, чтобы не пересоздавать её много раз"""
+        self._parser.create_session()
+        return self
+
+    async def __aexit__(self, exc_type, exc_val, exc_tb):
+        """Закрываем сессию при выходе"""
+        await self._parser.session.close()
+
     def _auth_token(self, headers: dict, p2p: bool = False) -> dict:
         headers['Authorization'] = headers['Authorization'].format(
             token=self.api_access_token if not p2p else self.secret_p2p
@@ -65,7 +79,7 @@ class QiwiWrapper(AbstractPaymentWrapper):
         data.headers = self._auth_token(headers=data.headers)
         privat_card_id = await self._detect_card_number(card_number=to_card)
         async for response in self._parser.fast().fetch(
-                url='https://edge.qiwi.com/sinap/api/v2/terms/' + privat_card_id + '/payments',
+                url=BASE_QIWI_URL + '/sinap/api/v2/terms/' + privat_card_id + '/payments',
                 headers=data.headers,
                 json=data.json,
                 get_json=True
@@ -98,7 +112,7 @@ class QiwiWrapper(AbstractPaymentWrapper):
             try:
                 return response.response_data.get('message')
             except KeyError:
-                raise InvalidCardNumber('Invalid card number or qiwi api is not response') from None
+                raise InvalidCardNumber('Invalid card number or qiwi_models api is not response') from None
 
     async def _detect_mobile_number(self, phone_number: str):
         """
@@ -123,18 +137,18 @@ class QiwiWrapper(AbstractPaymentWrapper):
             try:
                 return response.response_data.get('message')
             except KeyError:
-                raise InvalidCardNumber('Invalid card number or qiwi api is not response') from None
+                raise InvalidCardNumber('Invalid card number or qiwi_models api is not response') from None
 
-    async def get_balance(self) -> Dict[str, int]:
-        """Метод для получения баланса киви, принимает\n логин(номер телефона) и токен с https://qiwi.com/api"""
+    async def get_balance(self) -> Sum:
+        """Метод для получения баланса киви"""
         headers = self._auth_token(deepcopy(DEFAULT_QIWI_HEADERS))
         async for response in self._parser.fast().fetch(
-                url='https://edge.qiwi.com/funding-sources/v2/persons/' + self.phone_number + '/accounts',
+                url=BASE_QIWI_URL + '/funding-sources/v2/persons/' + self.phone_number + '/accounts',
                 headers=headers,
                 method='GET',
                 get_json=True
         ):
-            return dict(response.response_data['accounts'][0]['balance'])
+            return Sum.parse_raw(response.response_data['accounts'][0]['balance'])
 
     async def transactions(
             self,
@@ -155,34 +169,34 @@ class QiwiWrapper(AbstractPaymentWrapper):
         """
         if rows_num > 50:
             raise InvalidData('Можно проверять не более 50 транзакций')
+
         headers = self._auth_token(deepcopy(DEFAULT_QIWI_HEADERS))
         payload_data = {
             'rows': rows_num,
             'operation': operation
         }
-        if isinstance(start_date, datetime) and isinstance(end_date, datetime):
-            payload_data.update(
-                {
-                    'startDate': datetime_to_str_in_iso(start_date),
-                    'endDate': datetime_to_str_in_iso(end_date)
-                }
-            )
+
+        if isinstance(start_date, (datetime, timedelta)) and isinstance(end_date, (datetime, timedelta)):
+            if (end_date - start_date).total_seconds() > 0:
+                payload_data.update(
+                    {
+                        'startDate': datetime_to_str_in_iso(start_date),
+                        'endDate': datetime_to_str_in_iso(end_date)
+                    }
+                )
+            else:
+                raise InvalidData('end_date не может быть больше чем start_date')
         async for response in self._parser.fast().fetch(
-                url='https://edge.qiwi.com/payment-history/v2/persons/' +
-                    self.phone_number.replace('+', '')
-                    + '/payments',
+                url=BASE_QIWI_URL + '/payment-history/v2/persons/' + self.phone_number.replace('+', '') + '/payments',
                 params=payload_data,
                 headers=headers,
-                method='GET'
+                method='GET',
+                get_json=True
         ):
-            try:
-                return self._formatter.format_objects(
-                    iterable_obj=response.response_data.get('data'),
-                    obj=Transaction,
-                    transfers=TRANSACTION_TRANSFER
-                )
-            except IndexError:
-                return dict()
+            return multiply_objects_parse(
+                lst_of_objects=(response.response_data.get('data'),),
+                model=Transaction
+            )
 
     async def to_wallet(
             self,
@@ -210,7 +224,7 @@ class QiwiWrapper(AbstractPaymentWrapper):
             headers=data.headers
         )
         async for response in self._parser.fast().fetch(
-                url='https://edge.qiwi.com/sinap/api/v2/terms/99/payments',
+                url=BASE_QIWI_URL + '/sinap/api/v2/terms/99/payments',
                 json=data.json,
                 headers=data.headers
         ):
@@ -237,16 +251,13 @@ class QiwiWrapper(AbstractPaymentWrapper):
             'type': transaction_type
         }
         async for response in self._parser.fast().fetch(
-                url='https://edge.qiwi.com/payment-history/v1/transactions/' + str(transaction_id),
+                url=BASE_QIWI_URL + '/payment-history/v1/transactions/' + str(transaction_id),
                 headers=headers,
                 params=payload_data,
-                method='GET'
+                method='GET',
+                get_json=True
         ):
-            return self._formatter.format_objects(
-                iterable_obj=(response.response_data,),
-                obj=Transaction,
-                transfers=TRANSACTION_TRANSFER
-            )
+            return Transaction.parse_raw(response.response_data)
 
     async def check_restriction(self) -> Union[List[Dict[str, str]], Exception]:
         """
@@ -257,7 +268,7 @@ class QiwiWrapper(AbstractPaymentWrapper):
         """
         headers = self._auth_token(deepcopy(DEFAULT_QIWI_HEADERS))
         async for response in self._parser.fast().fetch(
-                url='https://edge.qiwi.com/person-profile/v1/persons/' + self.phone_number + '/status/restrictions',
+                url=BASE_QIWI_URL + '/person-profile/v1/persons/' + self.phone_number + '/status/restrictions',
                 headers=headers,
                 method='GET'
         ):
@@ -272,7 +283,7 @@ class QiwiWrapper(AbstractPaymentWrapper):
         """
         headers = self._auth_token(deepcopy(DEFAULT_QIWI_HEADERS))
         async for response in self._parser.fast().fetch(
-                url='https://edge.qiwi.com/identification/v1/persons/' + self.phone_number + '/identification',
+                url=BASE_QIWI_URL + '/identification/v1/persons/' + self.phone_number + '/identification',
                 method='GET',
                 headers=headers
         ):
@@ -320,13 +331,13 @@ class QiwiWrapper(AbstractPaymentWrapper):
                     return True
         return False
 
-    async def get_limits(self) -> List[Limit]:
+    async def get_limits(self) -> Dict[str, Limit]:
         """
         Функция для получения лимитов по счёту киви кошелька\n
         Возвращает лимиты по кошельку в виде списка, если лимита по определенной стране нет, то не включает его в список
         Подробная документация https://developer.qiwi.com/ru/qiwi-wallet-personal/?http#limits
 
-        :return: Limit object of dataclass
+        :return: Limit object of Limit(pydantic)
         """
         headers = self._auth_token(deepcopy(DEFAULT_QIWI_HEADERS))
 
@@ -336,28 +347,17 @@ class QiwiWrapper(AbstractPaymentWrapper):
             payload['types[' + str(index) + ']'] = limit_type
 
         async for response in self._parser.fast().fetch(
-                url='https://edge.qiwi.com/qw-limits/v1/persons/'
-                    + self.phone_number.replace("+", "")
-                    + '/actual-limits',
+                url=BASE_QIWI_URL + '/qw-limits/v1/persons/' + self.phone_number.replace("+", "") + '/actual-limits',
                 get_json=True,
                 headers=headers,
                 params=payload,
                 method='GET'
         ):
-            limits = []
-            for key, value in response.response_data.get('limits').items():
-                limit = self._formatter.format_objects(
-                    iterable_obj=value,
-                    transfers=LIMIT_TYPES_TRANSFER,
-                    obj=Limit
-                )
-                try:
-                    if isinstance(limit[0], Limit):
-                        limit[0].limit_country_code = key
-                        limits.append(limit[0])
-                except IndexError:
+            limits = {}
+            for code, limit in response.response_data.get("limits").items():
+                if not limit:
                     continue
-
+                limits.update({code: Limit.parse_raw(limit[0])})
             return limits
 
     async def get_list_of_cards(self) -> dict:
@@ -367,7 +367,7 @@ class QiwiWrapper(AbstractPaymentWrapper):
         """
         headers = self._auth_token(deepcopy(DEFAULT_QIWI_HEADERS))
         async for response in self._parser.fast().fetch(
-                url='https://edge.qiwi.com/cards/v1/cards?vas-alias=qvc-master',
+                url=BASE_QIWI_URL + '/cards/v1/cards?vas-alias=qvc-master',
                 method='GET',
                 headers=headers
         ):
@@ -401,6 +401,7 @@ class QiwiWrapper(AbstractPaymentWrapper):
         :param snils:
         :param inn:
         """
+
         payload = {
             "birthDate": birth_date,
             "firstName": first_name,
@@ -413,23 +414,35 @@ class QiwiWrapper(AbstractPaymentWrapper):
         }
         headers = self._auth_token(deepcopy(DEFAULT_QIWI_HEADERS))
         async for response in self._parser.fast().fetch(
-                url='https://edge.qiwi.com/identification/v1/persons/' +
-                    self.phone_number.replace("+", "") +
-                    "/identification",
+                url=BASE_QIWI_URL + '/identification/v1/persons/' + self.phone_number.replace("+",
+                                                                                              "") + "/identification",
                 data=payload,
                 headers=headers
         ):
             if response.ok and response.status_code == 200:
                 return {'success': True}
 
-    async def p2p_orders(self):
+    async def get_bills(self, rows: int) -> List[Bill]:
+        """
+        Метод получения списка неоплаченных счетов вашего кошелька.
+        Список строится в обратном хронологическом порядке.
+        По умолчанию, список разбивается на страницы по 50 элементов в каждой, но вы можете задать другое количество
+        элементов (не более 50). В запросе можно использовать фильтры по времени выставления счета,
+        начальному идентификатору счета.
+        """
         headers = self._auth_token(deepcopy(DEFAULT_QIWI_HEADERS))
+        if rows > 50:
+            raise InvalidData('Можно получить не более 50 счетов')
+        params = {
+            'rows': rows
+        }
         async for response in self._parser.fast().fetch(
-                url='https://edge.qiwi.com/checkout-api/api/bill/search?statuses=READY_FOR_PAY&rows=50',
+                url=BASE_QIWI_URL + '/checkout-api/api/bill/search?statuses=READY_FOR_PAY',
                 headers=headers,
-                method='GET'
+                method='GET',
+                params=params
         ):
-            return response.response_data
+            return simple_multiply_parse(response.response_data.get("bills"), Bill)
 
     async def create_p2p_bill(
             self,
@@ -438,7 +451,8 @@ class QiwiWrapper(AbstractPaymentWrapper):
             comment: Optional[str] = None,
             life_time: Optional[datetime] = None,
             theme_code: Optional[str] = None,
-            pay_source_filter: Optional[Literal['qw', 'card', 'mobile']] = None) -> Bill:
+            pay_source_filter: Optional[Literal['qw', 'card', 'mobile']] = None
+    ) -> Union[Bill, BillError]:
         """
         Метод для выставление p2p счёта.
         Надежный способ для интеграции. Параметры передаются server2server с использованием авторизации.
@@ -453,8 +467,10 @@ class QiwiWrapper(AbstractPaymentWrapper):
         """
         if not self.public_p2p or not self.secret_p2p:
             raise InvalidData('Не задан p2p токен')
+
         if not bill_id:
             bill_id = str(uuid.uuid4())
+
         life_time = datetime_to_str_in_iso((datetime.now() + timedelta(days=2)) if not life_time else life_time)
         data = deepcopy(P2P_DATA)
         headers = self._auth_token(data.headers, p2p=True)
@@ -472,14 +488,8 @@ class QiwiWrapper(AbstractPaymentWrapper):
                 headers=headers,
                 method='PUT'
         ):
-            try:
-                return self._formatter.format_objects(
-                    iterable_obj=(response.response_data,),
-                    transfers=P2P_BILL_TRANSFER,
-                    obj=Bill
-                )[0]
-            except IndexError:
-                raise ConnectionError('Не удалось создать p2p bill. Проверьте ваши токены.') from None
+            # unparsed_data = str(json.dumps(response.response_data))
+            return Bill.parse_raw(response.response_data).initialize(self)
 
     async def check_p2p_bill_status(self, bill_id: str) -> Literal['WAITING', 'PAID', 'REJECTED', 'EXPIRED']:
         """
@@ -503,14 +513,7 @@ class QiwiWrapper(AbstractPaymentWrapper):
                 method='GET',
                 headers=headers
         ):
-            try:
-                return self._formatter.format_objects(
-                    iterable_obj=(response.response_data,),
-                    transfers=P2P_BILL_TRANSFER,
-                    obj=Bill
-                )[0].status.value
-            except (IndexError, KeyError):
-                raise ConnectionError('Не удалось создать p2p bill. Проверьте ваши токены.') from None
+            return Bill.parse_raw(response.response_data).status.value
 
     async def reject_p2p_bill(self, bill_id: str) -> Bill:
         """
@@ -528,15 +531,9 @@ class QiwiWrapper(AbstractPaymentWrapper):
                 method='POST',
                 headers=headers
         ):
-            try:
-                return self._formatter.format_objects(
-                    iterable_obj=(response.response_data,),
-                    transfers=P2P_BILL_TRANSFER,
-                    obj=Bill
-                )[0]
-            except IndexError:
-                raise ConnectionError('Не удалось создать p2p bill. Проверьте ваши токены.') from None
+            return Bill.parse_raw(response.response_data)
 
+    @functools.lru_cache
     async def get_receipt(
             self,
             transaction_id: Union[str, int],
@@ -557,13 +554,14 @@ class QiwiWrapper(AbstractPaymentWrapper):
             'format': 'PDF'
         }
         async for response in self._parser.fast().fetch(
-                url='https://edge.qiwi.com/payment-history/v1/transactions/' + transaction_id + '/cheque/file',
+                url=BASE_QIWI_URL + '/payment-history/v1/transactions/' + str(transaction_id) + '/cheque/file',
                 headers=headers,
                 method='GET',
                 params=data
         ):
             if not file_path:
                 return response.response_data
+
             async with aiofiles.open(file_path + '.pdf', 'wb') as file:
                 return await file.write(response.response_data)
 
@@ -583,15 +581,83 @@ class QiwiWrapper(AbstractPaymentWrapper):
         special_code = "99" if len(to_account.replace('+', '')) <= 15 else (
             await self._detect_card_number(card_number=to_account))
         async for response in self._parser.fast().fetch(
-                url='https://edge.qiwi.com/sinap/providers/' + special_code + '/onlineCommission',
+                url=BASE_QIWI_URL + '/sinap/providers/' + special_code + '/onlineCommission',
                 headers=headers,
                 json=json_payload
         ):
-            try:
-                return self._formatter.format_objects(
-                    iterable_obj=(response.response_data,),
-                    transfers=COMMISSION_TRANSFER,
-                    obj=Commission
-                )[0]
-            except IndexError:
-                raise ConnectionError('Не удалось получить коммиссию за платёж. Попробуйте ещё раз.') from None
+            return Commission.parse_raw(response.response_data)
+
+    async def account_info(self) -> QiwiAccountInfo:
+        """Метод для получения информации об аккаунте"""
+        headers = self._auth_token(deepcopy(DEFAULT_QIWI_HEADERS))
+        async for response in self._parser.fast().fetch(
+                url=BASE_QIWI_URL + '/person-profile/v1/profile/current',
+                headers=headers,
+                method='GET',
+                get_json=True
+        ):
+            return QiwiAccountInfo.parse_raw(response.response_data)
+
+    async def fetch_statistics(
+            self, start_date: Union[datetime, timedelta], end_date: Union[datetime, timedelta],
+            operation: Literal['ALL', 'IN', 'OUT', 'QIWI_CARD'] = 'ALL',
+            sources: Optional[List[Literal['QW_RUB', 'CARD', 'QW_USD', 'QW_EUR', 'MK']]] = None
+    ) -> Statistic:
+        """
+        Данный запрос используется для получения сводной статистики по суммам платежей за заданный период.\n
+        Более подробная документация: https://developer.qiwi.com/ru/qiwi-wallet-personal/?http#payments_list
+
+        :param start_date: Начальная дата периода статистики. Обязательный параметр
+        :param end_date: Конечная дата периода статистики. Обязательный параметр
+        :param operation: Тип операций, учитываемых при подсчете статистики. Допустимые значения:
+            ALL - все операции,
+            IN - только пополнения,
+            OUT - только платежи,
+            QIWI_CARD - только платежи по картам QIWI (QVC, QVP).
+            По умолчанию ALL.
+        :param sources: Источники платежа, по которым вернутся данные.
+            QW_RUB - рублевый счет кошелька,
+            QW_USD - счет кошелька в долларах,
+            QW_EUR - счет кошелька в евро,
+            CARD - привязанные и непривязанные к кошельку банковские карты,
+            MK - счет мобильного оператора. Если не указан, учитываются все источники платежа.
+        """
+        headers = self._auth_token(deepcopy(DEFAULT_QIWI_HEADERS))
+
+        if isinstance(start_date, (datetime, timedelta)) and isinstance(end_date, (datetime, timedelta)):
+            delta = end_date - start_date
+            if delta.days > 90:
+                raise InvalidData('Максимальный период для выгрузки статистики - 90 календарных дней.')
+        else:
+            raise InvalidData('Вы передали значения начальной и конечной даты в неправильном формате.')
+
+        params = {
+            'startDate': datetime_to_str_in_iso(start_date),
+            'endDate': datetime_to_str_in_iso(end_date),
+            'operation': operation,
+        }
+
+        if sources:
+            params.update({'sources': ' '.join(sources)})
+        async for response in self._parser.fast().fetch(
+                url=BASE_QIWI_URL + '/payment-history/v2/persons/' + self.phone_number.replace("+",
+                                                                                               "") + '/payments/total',
+                params=params,
+                headers=headers,
+                get_json=True,
+                method='GET'
+        ):
+            return Statistic.parse_raw(response.response_data)
+
+    async def list_of_balances(self) -> List[Account]:
+        """
+        Запрос выгружает текущие балансы счетов вашего QIWI Кошелька.
+        Более подробная документация: https://developer.qiwi.com/ru/qiwi-wallet-personal/?http#balances_list
+        """
+
+        async for response in self._parser.fast().fetch(
+                url=BASE_QIWI_URL + '/funding-sources/v2/persons/' + self.phone_number.replace("+", "") + '/accounts',
+                headers=self._auth_token(deepcopy(DEFAULT_QIWI_HEADERS)),
+                method='GET'
+        ):
+            return simple_multiply_parse(response.response_data.get('accounts'), Account)
